/*
  freehandregion.cpp
  Copyright (C) 2008-2009 Tomohiro Masubuchi <bucchi@e-ml.jp>

  This file is part of qgmap.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; If not, see <http://www.gnu.org/licenses/>.
*/

//#define DEBUG

#include "qgmap.h"
#include "regionmap.h"
#include <stdio.h>

#define MAP_SIZE(width, height) (MAP_LINESIZE(width)*(height))
#define MAP_LINESIZE(width) (((width)+7)/8)
#define MAP_OFFSET(x,y,w) ((x)/8+(y)*MAP_LINESIZE(w))

RegionMap::RegionMap(int left, int top, int width, int height, int zoom)
{
	this->top = top;
	this->left = left;
	this->width = width;
	this->height = height;
	this->zoom = zoom;
	this->map = new unsigned char[MAP_SIZE(width,height)];
	this->shift = 0;
	memset(this->map, 0, MAP_SIZE(width,height));
}

RegionMap::RegionMap(const RegionMap &pos)
{
	top = pos.top;
	left = pos.left;
	width = pos.width;
	height = pos.height;
	zoom = pos.zoom;
	shift = pos.shift;
	map = new unsigned char[MAP_SIZE(width,height)];
	memcpy(map, pos.map, MAP_SIZE(width,height));
}


RegionMap::~RegionMap()
{
	delete [] map;
}

RegionMap RegionMap::zoomTo(int to_zoom)
{
	int new_width, new_height;
	int new_left, new_top;

	if(zoom == to_zoom){
		return RegionMap(*this);
	}

	if(zoom > to_zoom){
		// expand
		RegionMap new_map(*this);
		new_map.shift = zoom - to_zoom;
		return new_map;
	}

	// shrink
	int shift = to_zoom - zoom;
	new_left = left >> shift;
	new_top = top >> shift;
	new_width = width >> shift;
	new_height = height >> shift;
	if(new_width == 0)
		new_width = 1;
	if(new_height == 0)
		new_height = 1;

	RegionMap new_map(new_left, new_top, new_width, new_height, to_zoom);
	int x,y;

	for(y = 0; y < height; y++){
		for(x = 0; x < width; x++){
			if(is_set(x, y))
				new_map.set(x >> shift, y >> shift);
		}
	}
#ifdef DEBUG
	new_map.dump();
#endif
	return new_map;
}

QArray<QRect> RegionMap::getRects()
{
	int x,y;
	int w,h;
	QArray<QRect> rarray(0);
	int i;

	for(y = 0; y < height; y++){
		for(x = 0; x < width; x++){
			if(is_set(x, y)){
				w = scanWidth(x, y);
				h = scanHeight(x, y, w);
				i = rarray.size();
				rarray.resize(i + 1);
				rarray[i] = QRect((x + left) << shift, (y + top) << shift,
								  w << shift, h << shift);
			}
		}
	}
	return rarray;
}

int RegionMap::scanWidth(int x, int y)
{
	int xx;

	for(xx = x; xx < width; xx++){
		if(!is_set(xx, y))
			break;
		clear(xx, y);
	}
	return xx-x;
}

int RegionMap::scanHeight(int x, int y, int w)
{
	int xx, yy;
	int i;

	for(yy = y + 1; yy < height; yy++){
		for(xx = x; xx < x + w; xx++)
			if(!is_set(xx, yy))
				break;
		if(xx < x + w)
			break;

		for(i = x; i < x + w; i++)
			clear(i, yy);
	}
	return yy-y;
}

void RegionMap::set(int x, int y)
{
	unsigned char *p = map;
	p += MAP_OFFSET(x, y, width);
	*p |= 1 << (x & 7);
}

void RegionMap::clear(int x, int y)
{
	unsigned char *p = map;
	p += MAP_OFFSET(x, y, width);
	*p &= ~(1 << (x & 7));
}

bool RegionMap::is_set(int x, int y)
{
	unsigned char *p = map;
	p += MAP_OFFSET(x, y, width);
	return (*p) & (1 << (x & 7));
}

#ifdef DEBUG
void RegionMap::dump()
{
	printf("zoom=%d, shift=%d (%d,%d)-(%d,%d)\n", zoom, shift, left, top, left+width-1, top+height-1);
	unsigned char *p = map;
	int x,y;
	for(y=0;y<height;y++){
		for(x=0;x<width;x++){
			if(*p++)
				printf("8");
			else
				printf(".");
		}
		printf("\n");
	}
}
#endif
